import cachetools
import django.conf

from django.core.exceptions import PermissionDenied
from django.contrib.auth.models import User

from . import auth, models, rule


# Django creates an instance of Backend for every permission check. Since
# loading rules is somewhat expensive, we memoize the configuration items, as
# well as flush periodically to ensure we're using the latest rules.
@cachetools.cached(cache=cachetools.TTLCache(maxsize=128, ttl=60))
def cached_config(key):
    config = django.conf.settings.RBAC

    if key == "auth":
        return auth.load(config["AUTH"])

    if key == "rule":
        return rule.load(config["RULE"])

    raise Exception("cached_config: unknown key: %s" % key)


class Backend:
    def __init__(self, **kwargs):
        self.auth = kwargs.get("auth", cached_config("auth"))
        self.rule = kwargs.get("rule", cached_config("rule"))

    def authenticate(self, request, username=None, password=None):
        sn = self.auth.session()

        try:
            account = sn.authenticate(username, password)
        except Exception:
            raise PermissionDenied

        # Credentials are valid. Ensure the user object exists.
        user, _ = User.objects.get_or_create(username=username)

        # Determine roles from group membership.
        roles = set()

        for name, role in self.rule.roles.items():
            if sn.member(username, role.groups):
                roles.add(name)

        user.email      = account.email
        user.first_name = account.first_name
        user.last_name  = account.last_name

        if "_is_denied" in roles:
            roles.clear()

        user.is_active    = "_is_active" in roles
        user.is_staff     = "_is_staff"  in roles
        user.is_superuser = "_is_super"  in roles

        user.save()

        profile, _ = models.Profile.objects.get_or_create(user=user)

        profile.roles = sorted(list(roles))
        profile.save()

        # Close the session.
        sn.close()

        return user

    def get_user(self, user_id):
        user = User.objects.get(pk=user_id)

        # Ensure each user has a profile.
        if user and not hasattr(user, "profile"):
            models.Profile.objects.create(user=user)

        return user

    def has_perm(self, user_obj, perm, obj=None):
        import logging
        log = logging.getLogger('django')
        log.warn('backend has_perm: user = %s, perm = %s, obj = %s' % (user_obj, perm, obj))

        if hasattr(user_obj, "profile"):
            return self.rule.has_perm(user_obj.profile.roles, perm, obj)

        return False



import deepmerge
import functools
import importlib
import logging
import yaml

from inspect import getmembers, isfunction
from urllib  import parse, request
from wcmatch import fnmatch

from . import macros, middleware


try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader

log = logging.getLogger("netbox_rbac")

# Collect all public functions from macros.
functions = [
    (name, fn)
    for name, fn in getmembers(macros, isfunction)
    if not name.startswith("_")
]

# In order to guard against external dependencies, such as Gitlab, being
# unavailable, store the last configuration.
config = None


def load(paths):
    global config

    errors = []

    for path in paths:
        try:
            path = parse.urlparse(path, scheme="file")
            data = request.urlopen(path.geturl())

            config = Rule(yaml.load(data, Loader=Loader))

            return config

        except Exception as err:
            errors.append("%s: %s" % (path.geturl(), err))

    log.warn("load: no valid rules found: %s", errors)

    # Unable to load a new config, so return the current one.
    return config


class Rule:
    def __init__(self, config):
        self.roles = {}

        # Although YAML provides a mechanism for referencing one object from
        # another, it doesn't support deep merging, so we handle that manually.
        for name, role in config.items():
            for base in role.get("base", []):
                deepmerge.always_merger.merge(role, config[base])

            # Ignore template roles.
            if "groups" in role:
                self.roles[name] = Role(name, **role)

    # Given the user's roles, the requested permission, and the object, returns
    # whether or not the operation is allowed.
    def has_perm(self, roles, perm, obj):
        for role in roles:
            role = self.roles.get(role)

            # Permission is granted when:
            #  * The role is valid (defined in the configuration).
            #  * The requested permission is granted by the role.
            #  * The rule evaluates to True on the object.
            if role and role.has_perm(perm) and role.eval(obj):
                return True

        return False


class Role:
    def __init__(self, name, **kwargs):
        self.name    = name
        self.context = kwargs.get("context", {})
        self.groups  = kwargs.get("groups",  [])
        self.imports = kwargs.get("imports", [])
        self.perms   = kwargs.get("perms",   [])
        self.rule    = kwargs.get("rule")

        if self.rule:
            self.code = compile(self.rule, "<string>", "eval")
        else:
            self.code = None

    # Returns the result of evaluating the rule on the object, if both are
    # defined. Returns True otherwise.
    def eval(self, obj):
        if self.code and obj:
            context = {**self.context, "obj": obj}

            for name in self.imports:
                context[name] = importlib.import_module(name)

            for name, fn in functions:
                context[name] = functools.partial(fn, obj)

            context.update({
                "fnmatch": fnmatch.fnmatch,
                "request": middleware.request(),
            })

            return eval(self.code, context)

        return True

    # Returns whether or not this role provides the requested permission.
    def has_perm(self, perm):
        return fnmatch.fnmatch(perm, self.perms)



from netbox_rbac import rule

from netbox_rbac.tests.types import *


config = {
    # Grants all rack permissions when the rack is in DC1.
    "dcim_rack_admin_dc1": {
        "groups": ["DC1-Rack-Admins"],
        "perms": ["dcim.*_rack"],
        "rule": 'obj.site.name == "DC1"',
    },
    # Grants all rack change permission when the rack is in DC2.
    "dcim_rack_admin_dc2": {
        "groups": ["DC2-Rack-Users"],
        "perms": ["dcim.change_rack"],
        "rule": 'obj.site.name == "DC2"',
    },
    # Grants all permissions to all sites.
    "dcim_site_admin_all": {"groups": ["DCIM-Site-Admins"], "perms": ["dcim.*_site"]},
    "import_test": {
        "groups": ["DCIM-Site-Admins"],
        "imports": ["os", "os.path"],
        "perms": ["test.*_import"],
        "rule": 'os.path.basename("/foo/bar") == "bar"',
    },
}


class TestRule:
    def test_rule(self):
        roles = set(config.keys())
        rules = rule.Rule(config)

        tests = [
            # Blank rack, granted by dcim_rack_admin_dc1.
            ("dcim.add_rack", True, None),
            # Blank region, no roles grant region permissions.
            ("dcim.add_region", False, None),
            # DC1 rack, all permissions granted.
            ("dcim.add_rack", True, Rack(name="R1", site=Site(name="DC1"))),
            ("dcim.change_rack", True, Rack(name="R1", site=Site(name="DC1"))),
            ("dcim.delete_rack", True, Rack(name="R1", site=Site(name="DC1"))),
            # All site permissions granted.
            ("dcim.add_site", True, None),
            ("dcim.add_site", True, Site(name="DC1")),
            ("dcim.change_site", True, Site(name="DC1")),
            ("dcim.delete_site", True, Site(name="DC1")),
            # DC2 rack, only changes are allowed.
            ("dcim.add_rack", False, Rack(name="R1", site=Site(name="DC2"))),
            ("dcim.change_rack", True, Rack(name="R1", site=Site(name="DC2"))),
            ("dcim.delete_rack", False, Rack(name="R1", site=Site(name="DC2"))),
            # Module import test.
            ("test.test_import", True, Site(name="DC1")),
        ]

        for perm, result, obj in tests:
            assert rules.has_perm(roles, perm, obj) == result



import collections

from . import ldap, mock


Account = collections.namedtuple("Account", ["email", "first_name", "last_name"])


def load(config):
    if "LDAP" in config:
        return ldap.Driver(config["LDAP"])

    if "MOCK" in config:
        return mock.Driver(config["MOCK"])

    raise Exception("no auth driver configured")



from django.shortcuts import render
from django.views.generic import View

from .backend import Backend


class Roles(View):
    def get(self, request):
        backend = Backend()
        context = {
            "rule": backend.rule,
            "user": request.user,
        }

        return render(request, "netbox_rbac/roles.html", context)



from .. import auth


class Driver:
    def __init__(self, config):
        self.config = config

    def session(self):
        return Session(**self.config)


class Session:
    def __init__(self, users):
        self.users  = users
        self.groups = []

    def authenticate(self, username, password):
        user = filter(
            lambda u: u["username"] == username and u["password"] == password,
            self.users,
        )

        user = next(user, None)

        if not user:
            raise Exception("invalid credentials")

        self.groups = user["groups"]

        return auth.Account(
            email      = user["email"],
            first_name = user["first_name"],
            last_name  = user["last_name"],
        )

    def close(self):
        pass

    def member(self, username, groups):
        for group in groups:
            if group in self.groups:
                return True

        return False



from django.core.exceptions import PermissionDenied
from netbox.views           import generic
from utilities.permissions  import resolve_permission


#
# Bulk view mixins.
#
class BulkCheckView:
    def post(self, request, *args, **kwargs):
        perm = self.get_required_permission()

        if request.POST.get('_all'):
            qs = self.queryset.all()

            if self.filterset:
                qs = self.filterset(request.GET, qs).qs

        else:
            qs = self.queryset.filter(pk__in=request.POST.getlist('pk'))

        for obj in qs.iterator():
            if not request.user.has_perm(perm, obj):
                raise PermissionDenied('%s %s' % (resolve_permission(perm)[1], obj))

        return super().post(request, *args, **kwargs)


class BulkDeleteView(BulkCheckView, generic.BulkDeleteView):
    pass


class BulkEditView(BulkCheckView, generic.BulkEditView):
    pass


class BulkRenameView(BulkCheckView, generic.BulkRenameView):
    pass


#
# Object view mixins.
#
class ObjectCheckView:
    def get(self, request, *args, **kwargs):
        obj  = self.get_object(kwargs)
        perm = self.get_required_permission()

        if not request.user.has_perm(perm, obj):
            raise PermissionDenied('%s %s' % (resolve_permission(perm)[1], obj))

        return super().get(request, *args, **kwargs)


class ObjectDeleteView(ObjectCheckView, generic.ObjectDeleteView):
    pass


class ObjectEditView(ObjectCheckView, generic.ObjectEditView):
    pass
